#include <stdio.h>
#include <pthread.h>
#include <stdlib.h>
#define _GNU_SOURCE
#include <fcntl.h>
#include <unistd.h>
#include <netinet/in.h>
#include <sys/epoll.h>
#include <string.h> 
#include "KVMessageFormat.h"
#define MAX_EVENTS 10


void *worker(void *args) {
    struct epoll_event ev,events[MAX_EVENTS];
    struct epoll_event socketEvent;
    socketEvent.events=EPOLLIN;

    int read_pipe, conn_sock, nfds, i;
    int epollfd;
    int newfd;
    
    epollfd = epoll_create1(0);
    if (epollfd == -1) {
        perror("epoll_create1");
        exit(EXIT_FAILURE);
    }

    /* Add the pipe which we get from main thread */
    read_pipe = ((int *) args)[0];   
    ev.events = EPOLLIN;
    ev.data.fd = read_pipe;
    if (epoll_ctl(epollfd, EPOLL_CTL_ADD, read_pipe, &ev) == -1) {
        perror("epoll_ctl: read_pipe");
        exit(EXIT_FAILURE);
    }

    while (1) {
        nfds = epoll_wait(epollfd, events, MAX_EVENTS, -1);
        if (nfds == -1) {
            perror("epoll_wait");
            exit(EXIT_FAILURE);
        }
        for ( i= 0; i < nfds; ++i ) {
            if (events[i].data.fd == read_pipe) {
                /*  if we get a request from main thread to add new client */
                read(read_pipe, &newfd, sizeof(newfd));
              
                printf("\nread %d\n", read_pipe);
                ev.data.fd=newfd;
                ev.events = EPOLLIN | EPOLLRDHUP;
                if (epoll_ctl(epollfd, EPOLL_CTL_ADD, newfd,  &ev) == -1) {
                    perror("epoll_ctl: read_pipe");
                    exit(EXIT_FAILURE);
                }

               
            } else {
                /*  if we get a request from client (GET, PUSH, DEL) */
                int flag = events[i].events;

                if (flag & EPOLLRDHUP) {
                    /* Connection was closed. */
                    epoll_ctl( epollfd, EPOLL_CTL_DEL, events[i].data.fd , NULL );
                    close(events[i].data.fd);
                    continue; 
                }

                if (flag & EPOLLIN) {
                    /* Parse the actual message from client */
                    struct message *requestMessage= malloc(sizeof(struct message));
                    int readlength=read(events[i].data.fd  , requestMessage, sizeof(struct message)); 
                    printf("[Message Received from client] (%d, %s, %s)\n",requestMessage->status,requestMessage->key,requestMessage->value);               
                    /*Process the message and send appropriate response to client*/
                    requestMessage->status=200;
                    write(events[i].data.fd,requestMessage, sizeof(struct message));
                    fflush(stdout);
                }                                 
               
            }
        }
    }

    
}


int main (int argc, int argv) {

    int i;
    int next_thread_to_assign = 0;                                   
    pthread_t *worker_threads;                             
    int pool_thread_size = 5;                       // TODO: get pool thread size from config file
    int sockfd, newsockfd, portno, clilen, n;
    struct sockaddr_in serv_addr, cli_addr; 

    //int *pipes = (int*) malloc(pool_thread_size * 2 * sizeof(pipes));
    int pipes[5][2];                                // TODO: initialize pipes dynamically

    worker_threads = malloc(pool_thread_size * sizeof(pthread_t));
    for( i=0; i < pool_thread_size; i++ ) {
        pipe(pipes[i]);
        pthread_create( &worker_threads[i], NULL, worker, &pipes[i]);
    }

    sockfd = socket(AF_INET, SOCK_STREAM, 0); 
    if (sockfd < 0) {
        perror("ERROR opening socket");
        exit(1);
    }
    memset(&serv_addr, 0, sizeof(serv_addr));
    portno = 8000;
    serv_addr.sin_family = AF_INET;
    serv_addr.sin_port = htons(portno);
    serv_addr.sin_addr.s_addr = INADDR_ANY;
    int option =1;
    setsockopt(sockfd, SOL_SOCKET, SO_REUSEADDR, &option, sizeof(option));
    if (bind(sockfd, (struct sockaddr *) &serv_addr, sizeof(serv_addr)) < 0) {
        perror("ERROR on binding");
        exit(1);
    }
    
    listen(sockfd, 5);
    clilen = sizeof(cli_addr);

    while(1) {

        fflush(stdout);
        newsockfd = accept(sockfd, (struct sockaddr *) &cli_addr, &clilen);
        if (newsockfd < 0) 
            perror("ERROR on accept");

        write(pipes[next_thread_to_assign][1], &newsockfd, sizeof(newsockfd));
        
        next_thread_to_assign = (next_thread_to_assign+1) % pool_thread_size;
    } 

    return 0;
}
